% *** FOD Algorithm ***
% *VB algorithm with first order difference (FOD) prior
%  to solve set of equations y = Mx + n, where x is hight-resolution image
% *see paper: J. evk, V. mdl and F. roubek, An Adaptive Correlated
% Image Prior for Image Restoration Problems, IEEE signal processing
% letters, 2018.
%
% INPUT:
%  K ....... number of low-resolution images
%  yk ...... cell array of K low-resolution (LR) images (cell 1 x K)
%             each cell contains one LR images (matrix of size p1 x p2)
%  Mk ...... cell arary of K degradation matrix  (cell 1 x K)
%             each cell contains one degradation matrix (of size p1*p2 x N)
%  n1,n2 ... size of high-resolution image (n1 x n2: important for visulation)
%
% written by Jakub Sevcik, Vaclav Smidl and Filip Sroubek 2018
% for more information contact: jsevcik@rice.zcu.cz

%% data preparing
load data_example

M = [];
y = [];
for i = 1:K
    M = [M;Mk{i}];
    y = [y;yk{i}(:)];
end

[p, N] = size(M);
[p1, p2] = size(yk{1});

%% preparing of auxiliary structures
Dy = sparse( diag(ones(n1-1,1)) ) - sparse( diag(ones(n1-2,1),1) );
Dy(n1-1,n1) = -1;
Dy = sparse( kron(eye(n2),Dy) );

Dx = sparse(kron(sparse([eye(n2-1) zeros(n2-1,1)]),sparse(eye(n1))));
Dx = sparse(Dx - sparse(  kron(sparse( [zeros(n2-1,1) eye(n2-1)] ), sparse(eye(n1))) ));

Diff = [Dy; Dx];

for i = 1:1:K
    MktMk{i} = (Mk{i})'*Mk{i};
end

%% inicialization of parameters and hyper-parameters
hat_alpha = 1e-5*ones(size(Diff,1),1);
hat_beta = ones(K,1);

gamma_0_alpha = 1e-8;
delta_0_alpha = 1e-8;

gamma_0_beta = 1e-8;
delta_0_beta = 1e-8;

iterations = 100;

%% computation
for iter = 1:iterations
    % ------ x ------
    sigma_x = inv(  M'*sparse(diag(kron(hat_beta,ones(p1*p2,1))))*M + ...
                    Diff'*sparse(diag(hat_alpha))*Diff  );
        
    hat_x = sigma_x*( M'*(kron(hat_beta,ones(p1*p2,1)).*y) );
    
    hat_x(hat_x<0) = 0;
    hat_xxt = hat_x*hat_x' + sigma_x;   
        
    % ------ alpha ------
    gamma_alpha = gamma_0_alpha + 0.5;
    delta_alpha = delta_0_alpha + 0.5*(Diff*hat_x).^2 + 0.5*diag(sparse(Diff)*sigma_x*sparse(Diff)');
    
    hat_alpha = gamma_alpha./delta_alpha;

    % ------ beta ------ 
    gamma_beta = gamma_0_beta + 0.5*(p1*p2);
    for i = 1:K
        delta_beta = delta_0_beta + 0.5*sum(sum(hat_xxt.*MktMk{i})) - (yk{i}(:))'*Mk{i}*hat_x + 0.5*(yk{i}(:))'*yk{i}(:);
        hat_beta(i) = gamma_beta / delta_beta;
    end

    if mod(iter,10) == 0
        disp(['FOD alg.: iteration ' num2str(iter) '/' num2str(iterations)])    
    end
end

%% visualisation of the reconstruction
figure
imagesc(reshape(hat_x,n1,n2))
title('reconstruction of high-resolution image')
colormap gray
axis off equal