% *** TV Algorithm ***
% *VB algorithm with TV prior
%  to solve set of equations y = Mx + n, where x is hight-resolution image
% *see paper: J. evk, V. mdl and F. roubek, An Adaptive Correlated
% Image Prior for Image Restoration Problems, IEEE signal processing
% letters, 2018.
%  
% INPUT:
%  K ....... number of low-resolution images
%  yk ...... cell array of K low-resolution (LR) images (cell 1 x K)
%             each cell contains one LR images (matrix of size p1 x p2)
%  Mk ...... cell arary of K degradation matrix  (cell 1 x K)
%             each cell contains one degradation matrix (of size p1*p2 x N)
%  n1,n2 ... size of high-resolution image (n1 x n2: important for visulation)
%
%  eps ..... optional parameter for better regularization, default value
%            eps = 0, in case of numerically close to singular matrix sigma_x
%            set small positive value (e.g. 1e-12)
%
% written by Jakub Sevcik, Vaclav Smidl and Filip Sroubek 2018
% for more information contact: jsevcik@rice.zcu.cz

%% data preparing
eps = 0; %see script header
load data_example

M = [];
y = [];
for i = 1:K
    M = [M;Mk{i}];
    y = [y;yk{i}(:)];
end

[p, N] = size(M);
[p1, p2] = size(yk{1});

%% preparing of auxiliary structures
Dy = sparse(diag(ones(n1-1,1))) - sparse(diag(ones(n1-2,1),1));
Dy(n1-1,n1) = -1;
Dy = sparse(kron(eye(n2-1),Dy));
Dy = [Dy zeros((n1-1)*(n2-1),n1)];

Dx = sparse(kron(sparse([eye(n2-1) zeros(n2-1,1)]),sparse(eye(n1))));
Dx = sparse(Dx - sparse(kron(sparse([zeros(n2-1,1) eye(n2-1)]), sparse(eye(n1)))));
for i=(n2-1):-1:1
    Dx(n1*i,:) = [];
end

Diff = [Dy; Dx];

for i = 1:1:K
    MktMk{i} = (Mk{i})'*Mk{i};
end

%% inicialization of parameters and hyper-parameters
hat_alpha = 1e-5;
hat_beta = ones(K,1);

w = 1*ones((n1-1)*(n2-1),1); 

gamma_0_alpha = 1e-8;
delta_0_alpha = 1e-8;

gamma_0_beta = 1e-8;
delta_0_beta = 1e-8;

iterations = 100;

%% computation
for iter = 1:iterations
    % ------ x ------
    sigma_x = inv(  M'*sparse(diag(kron(hat_beta,ones(p1*p2,1))))*M + ...
                    hat_alpha*Diff'*sparse(diag(kron([1;1],1./sqrt(w))))*Diff + ...
                    eps*speye(N));

    hat_x = sigma_x*( M'*(kron(hat_beta,ones(p1*p2,1)).*y) );
    
    hat_x(hat_x<0) = 0;     % positivity of x
    hat_xxt = hat_x*hat_x' + sigma_x;   
    
    % ------ w ------
    w = (Dy*hat_x).^2 + (Dx*hat_x).^2 + diag(Dy*sigma_x*Dy') + diag(Dx*sigma_x*Dx');

    % ------ alpha ------
    gamma_alpha = gamma_0_alpha + 0.5*(n1*n2);
    delta_alpha = delta_0_alpha + sum(sqrt(w));
    
    hat_alpha = gamma_alpha / delta_alpha;
    
    % ------ beta ------ 
    gamma_beta = gamma_0_beta + 0.5*(p1*p2);
    for i = 1:K
        delta_beta = delta_0_beta + 0.5*sum(sum(hat_xxt.*MktMk{i})) - (yk{i}(:))'*Mk{i}*hat_x + 0.5*(yk{i}(:))'*yk{i}(:);
        hat_beta(i) = gamma_beta / delta_beta;
    end

    if mod(iter,10) == 0
        disp(['TV alg.: iteration ' num2str(iter) '/' num2str(iterations)])    
    end
end

%% visualisation of the reconstruction
figure
imagesc(reshape(hat_x,n1,n2))
title('reconstruction of high-resolution image')
colormap gray
axis off equal
